/**************************************************************************************
 * \file
 * ClasserTester - Implementation of Classer API Use Cases
 *                 See companion Word document "Classer API Use Cases"
 *
 * Copyleft (C) - Boston University CELEST Technology Laboratory
 *************************************************************************************/

#include "Logger.h"
#include "Classifier.h"
#include "Image.h"
#include "NameList.h"
#include "Palette.h"

/*************************************************************************************/
// The meta-data:
int imageWidth  = 360;
int imageHeight = 600;
int numFeatures = 41;
int numClasses  = 0; // Set in UC_0_defineClassesAndFeatures()
int numSamples  = imageWidth * imageHeight;

/*************************************************************************************/
// Persistent objects
Image *     bosImg        = 0; /* The data set of interest, in Image form           */
Data  *     bosGt         = 0; /* Ground truth subset of the image data             */
NameList *  classNames    = 0; /* More meta-data - the names of the class labels    */
NameList *  dimNames      = 0; /* The names of the available input features         */
float *     dataArray     = 0; /* The raw data, read from file                      */
int   *     labelArray    = 0; /* The sample labels                                 */
DimList *   dimList       = 0; /* Object to contain the dimNames list               */
Classifier *clsAllGt      = 0; /* A classifier trained w/ all the ground truth      */
ViewSet *   smallSubset   = 0; /* A five-sample subset of the Boston data           */

// A place to store all the timestamps
char ts[128];

/*************************************************************************************/
// Input/output files:
char *rawFeaturedataFname  = "bos_data_ErdasExport.asc";
char *rawSampleLabelsFname = "bos_gt_ErdasExport.asc"; 
char *bostonImgFname_bin   = "bostonImage.bin";
char *bostonImgFname_txt   = "bostonImage.txt";
char *bostonGt_Fname_bin   = "bostonGt.bin";
char *bostonGt_Fname_txt   = "bostonGt.txt";
char *thematicMapFname     = "bostonThematic.ppm";
char *confidenceMapFname   = "bostonConf.pgm";
char *confusionMatrixFname = "bostonCmat.pgm";
char *decisionSurfFname    = "decisionSurface.ppm";
char *bostonLegendFname    = "bostonLegend.ppm";
char *outputClassFname     = "outputClasses.txt";
char *outputVectorFname    = "outputVector.txt";

/*************************************************************************************
 Generates a simple DimList object for use with this testbed
 */
DimList *genFeatureRange (char *name, char *prefix, int low, int high) {
	NameList *featureRange = new NameList ();
	for (int i = low; i <= high; i++) {
		ostringstream dimName;
		dimName << prefix << i;
		featureRange->add(dimName.str());
	}

	ostringstream dimsOss;
	dimsOss << featureRange;
	DimList *dimList = new DimList (name, *featureRange);

	return dimList;
}

/*************************************************************************************
 Define the names for the classes and features to use with this data set.
 */
void UC_0_defineClassesAndFeatures() {

	// Class label names are givens
	classNames = new NameList ();
	classNames->add("unlab.");
	classNames->add("miss");
	classNames->add("beach");
	classNames->add("ocean");
	classNames->add("ice");
	classNames->add("river");
	classNames->add("road");
	classNames->add("park");
	classNames->add("resid"); 
	classNames->add("indst");
	numClasses = classNames->size();

									INDENT LOG ("Defining class names: " << *classNames << " and " << numFeatures << " feature names\n");

	// Use generic names for the features
	dimNames = new NameList ();
	for (int i = 1; i <= numFeatures; i++) {
		ostringstream dimName;
		dimName << "d" << i;
		dimNames->add(dimName.str());
	}

	dimList = new DimList ("AllFeatures", *dimNames);
}

/*************************************************************************************/
void UC_0_readDataValues(char *imgFname) {
	ifstream *imgIfs = new ifstream (imgFname, ios::in);
	if (!(*imgIfs)) MSG_EXCEPTION ("UC_0_readDataValues() - Failed to open Erdas input file " << imgFname);

									INDENT LOG ("Note - '.' = 10000 entries read\n");
									INDENT LOG ("Reading data values: ");

	string line, bitBucket;
	int entryNum = 0;
	float dataValue;
	int colIdx;
	// Allocate space to store the data
	dataArray = new float[numSamples * numFeatures];

	// Discard four header lines
	for (int i = 0; i < 4; i++) {
		getline(*imgIfs, bitBucket);
	}

	// Discard first two columns - we don't need the pixel coordinates	
	while ((*imgIfs) >> bitBucket) { (*imgIfs) >> bitBucket;
		colIdx = 0;
		for (int i = 0; i < numFeatures; i++) {
			*imgIfs >> dataValue;
      dataArray[entryNum*numFeatures + i] = dataValue;
		}
		// Print a heartbeat every 10000 entries read
		if ((entryNum++ > 0) && (entryNum % 10000 == 0)) { cout << "."; cout.flush(); }	
	}

									LOG ("\n");

	imgIfs->close();
}

/*************************************************************************************/
void UC_0_readDataLabels(char *labelFname) {
	ifstream *labelIfs = new ifstream (labelFname, ios::in);
	if (!(*labelIfs)) MSG_EXCEPTION ("UC_0_readDataLabels() - Failed to open Erdas label file " << labelFname);

									INDENT LOG ("Reading data labels: ");

	string line, bitBucket;
	int entryNum = 0;
	int dataLabel;
	// Allocate space to store the data
	labelArray = new int[numSamples];

	// Discard four header lines
	for (int i = 0; i < 4; i++) {
		getline(*labelIfs, bitBucket);
	}

	// Discard first two columns - we don't need the pixel coordinates	
	while ((*labelIfs) >> bitBucket) { (*labelIfs) >> bitBucket;
		*labelIfs >> dataLabel;
    labelArray[entryNum] = dataLabel;
		// Print a heartbeat every 10000 entries read
		if ((entryNum++ > 0) && (entryNum % 10000 == 0)) { cout << "."; cout.flush(); }	
	}

									LOG ("\n");

	labelIfs->close();
}

/*************************************************************************************
 Do all the preprocessing for this data set. Assume the
 starting point is the export from an ERDAS GIS:
  - One file with the feature information for the entire image
	  - One row per image pixel
		- One column per image feature
  - One file with the labels for all the image pixels
	  - Each "ground-truth" pixel specifies a label defined by
		  an integer index.
		- Each non-ground-truth pixel has a label index of 0
 Image details are hardwired in this example. In a real
 application they would be passed in from the caller.

 Pre-conditions: 
	    The names of the classes and features for the data set, and
	    Either - A file containing feature data for N samples
			         A file containing labels for N samples
		  Or     - An already-preprocessed binary data set

 Post-conditions: h
	    A processed binary data set.
 */
void UC_0_preprocess() {
									LOG_ENTER (0, "\n");
									INDENT LOG ("************************************************************************************");  timeStamp(ts); LOG (ts << endl);
									INDENT LOG ("Use Case 0 - Preprocess data:\n");
									INDENT LOG ("----------------------------\n");

	bosImg = new Image ("bostonImage", imageWidth, imageHeight, classNames, dimNames);

									INDENT LOG ("Preprocessing raw data\n");
	// Strip off ERDAS-specific headers and pixel coordinates,
	// yielding files with raw feature and label data.
	UC_0_readDataValues (rawFeaturedataFname);
	UC_0_readDataLabels (rawSampleLabelsFname);

	bosImg->setData(dataArray);
	bosImg->setLabels(labelArray);

									INDENT LOG ("Normalizing feature data\n");

	bosImg->normalize();

									INDENT LOG ("Saving Image in binary format -> " << bostonImgFname_bin << " (will use this the next time ClasserTester is run)\n");

	ofstream *imgBinOfs = new ofstream (bostonImgFname_bin, ios::out | ios::binary);
	if (!(*imgBinOfs)) MSG_EXCEPTION ("UC_0_all() - Failed to open image output file (binary version) " << bostonImgFname_bin);
	bosImg->save(*imgBinOfs, false, true, false);
	
									INDENT LOG ("Saving Image in text format   -> " << bostonImgFname_txt << " (inspect this file to see input/output format)\n");
	ofstream *imgTxtOfs = new ofstream (bostonImgFname_txt, ios::out);
	if (!(*imgTxtOfs)) MSG_EXCEPTION ("UC_0_all() - Failed to open image output file (text version) " << bostonImgFname_bin);
	bosImg->save(*imgTxtOfs, true, true, true);
									INDENT LOG ("Data preprocessing complete.\n\n");

	delete bosImg;

									LOG_LEAVE(0);
}

/*************************************************************************************/
void UC_1_loadImage() {
									LOG_ENTER (0, "\n");
									INDENT LOG ("************************************************************************************");  timeStamp(ts); LOG (ts << endl);
									INDENT LOG ("Use case 1 - Loading binary image data:\n");
									INDENT LOG ("--------------------------------------\n");

	ifstream *imgBinIfs = new ifstream (bostonImgFname_bin, ios::in | ios::binary);
	bosImg = new Image ("bostonImage", *imgBinIfs, imageWidth, imageHeight, classNames, dimNames);

									INDENT LOG ("Binary data loaded\n\n");

									INDENT LOG ("Plotting data in dimensions 1 and 2:\n");
	bosImg->plot(cout,0,1,true);
									INDENT LOG ("\n");
									INDENT LOG ("Histogram of data by class:\n\n");
	bosImg->hist(cout);

									LOG_LEAVE (0);
}

/*************************************************************************************/
void UC_2_defineGroundTruth() {
									LOG_ENTER (0, "\n");
									INDENT LOG ("************************************************************************************");  timeStamp(ts); LOG (ts << endl);
									INDENT LOG ("Use case 2 - Splitting out the ground truth data:\n");
									INDENT LOG ("--------------------------------------\n");

									INDENT LOG ("Copying data set to form ground truth set\n");
	bosGt = new Data ("bostonGT", bosImg);  // Use copy constructor to copy the image data set

									INDENT LOG ("Removing unlabeled samples from ground truth set...");
	bosGt->deleteClass (0);
	LOG (" Done.\n\n");

									INDENT LOG ("Histogram of ground truth data by class:\n\n");
	bosGt->hist (cout);

									LOG ("\n");
									INDENT LOG ("Saving ground truth in binary format -> " << bostonGt_Fname_bin << "\n");

	ofstream *gt_BinOfs = new ofstream (bostonGt_Fname_bin, ios::out | ios::binary);
	if (!(*gt_BinOfs)) MSG_EXCEPTION ("UC_0_all() - Failed to open ground truth output file (binary version) " << bostonGt_Fname_bin);
	bosGt->save(*gt_BinOfs, false, true, false);
	
									INDENT LOG ("Saving ground truth in text format   -> " << bostonGt_Fname_txt << "\n");
	ofstream *gt_TxtOfs = new ofstream (bostonGt_Fname_txt, ios::out);
	if (!(*gt_TxtOfs)) MSG_EXCEPTION ("UC_0_all() - Failed to open ground truth output file (text version) " << bostonGt_Fname_bin);
	bosGt->save(*gt_TxtOfs, true, true, true);

									LOG ("\n");
									INDENT LOG ("Ground truth extraction complete.\n\n");

																	LOG_LEAVE (0);
}

/*************************************************************************************/
void UC_3_trainTest () {
									LOG_ENTER (0, "\n");
									INDENT LOG ("************************************************************************************");  timeStamp(ts); LOG (ts << endl);
									INDENT LOG ("Use case 3 - Train with one subset of ground truth, test on another:\n");
									INDENT LOG ("-------------------------------------------------------------------\n");

	// Use a ViewSet object to automate the splitting and generation of orderings
	ViewSet *vs = new ViewSet ("bosViewSet", bosGt, 2, 5);
	View *v0 = vs->getView(0);
	View *v1 = vs->getView(1);

	//Create a couple more ViewSets to hold the training and testing views.
	ViewSet *vsTr = new ViewSet ("Training");
	vsTr->addView(v0);
	ViewSet *vsTe = new ViewSet ("Testing");
	vsTe->addView(v1);

	Classifier *cls = new Classifier (*(new string ("Artmap")), dimList, 5, *classNames);

									INDENT LOG ("Training with half the data\n");
	cls->train(vsTr);
									INDENT LOG ("Training completed in " << cls->getTrainTime() << " seconds\n");
	cls->test(vsTe);
									INDENT LOG ("Testing completed in " << cls->getTestTime() << " seconds, " << cls->getPercentCorrect() << "% correct\n");
									INDENT LOG ("ARTMAP category nodes required: ");

	int num = cls->getNumClassifiers();

	for (int i = 0; i < num; i++) { LOG (cls->getClassifier(i)->getInt("f2Nodes") << " "); }
									LOG ("\n");

	ConfusionMatrix *cm = cls->getConfusionMatrix();
	cout << *(cm->charPlot());

	delete vs;
	delete v0;
	delete v1;
	delete cls;
	delete vsTr;
	delete vsTe;
	delete cm;

									LOG_LEAVE (0);
}

/*************************************************************************************/
void UC_4_CVtrainTest () {
																	LOG_ENTER (0, "\n");
									INDENT LOG ("************************************************************************************");  timeStamp(ts); LOG (ts << endl);
									INDENT LOG ("Use case 4 - Cross-validated training/testing:\n");
									INDENT LOG ("---------------------------------------------\n");

	ViewSet *vs = new ViewSet ("bosViewSet", bosGt, 5, 5);

	Classifier *cls = new Classifier (*(new string ("Artmap")), dimList, 5, *classNames);
	cls->setHeartbeat(false);
									INDENT LOG ("Training (cross-validated)\n");
	cls->CvTrain(vs);
									INDENT LOG ("Training completed in " << cls->getTrainTime() << " seconds\n");
	cls->CvTest(vs);
									INDENT LOG ("Testing completed in " << cls->getTestTime() << " seconds, " << cls->getPercentCorrect() << "% correct\n");

	ConfusionMatrix *cm = cls->getConfusionMatrix();
	cout << *(cm->charPlot());

	delete vs;
	delete cls;
	delete cm;

									LOG_LEAVE (0);
}

/*************************************************************************************/
void UC_5_trainAllGt () {
									LOG_ENTER (0, "\n");
									INDENT LOG ("************************************************************************************");  timeStamp(ts); LOG (ts << endl);
									INDENT LOG ("Use case 5 - Train five classifier instances with all of the ground truth:\n");
									INDENT LOG ("-------------------------------------------------------------------------\n");

	ViewSet *vs = new ViewSet ("bosViewSet", bosGt, 1, 5);

	clsAllGt = new Classifier (*(new string ("Artmap")), dimList, 5, *classNames);

									INDENT LOG ("Training with half the data\n");
	clsAllGt->train(vs);

									LOG_LEAVE (0);
}

/*************************************************************************************/
void UC_6_genThematicMap () {
									LOG_ENTER (0, "\n");
									INDENT LOG ("************************************************************************************");  timeStamp(ts); LOG (ts << endl);
									INDENT LOG ("Use case 6 - Generate a thematic map with the trained classifier:\n");
									INDENT LOG ("----------------------------------------------------------------\n");

	ofstream *ofsThematicMap = new ofstream (thematicMapFname, ios::out | ios::binary);
	if (!(*ofsThematicMap)) MSG_EXCEPTION ("UC_6_genThematicMap() - Failed to open thematic map output file " << thematicMapFname);
	ofstream *ofsConfidenceMap = new ofstream (confidenceMapFname, ios::out | ios::binary);
	if (!(*ofsConfidenceMap)) MSG_EXCEPTION ("UC_6_genThematicMap() - Failed to open confidence map output file " << confidenceMapFname);

									LOG ("\n");
									INDENT LOG ("Generating thematic map and confidence map for Boston data set\n");
	// Use the image loaded in UC_1_loadImage(), and the classifier trained in UC_5_trainAllGt()
	clsAllGt->thematicPpm(*ofsThematicMap, *bosImg, *ofsConfidenceMap);
									LOG ("\n");
									INDENT LOG ("Done.\n\n");
									INDENT LOG ("Finished maps:\n");
									INDENT LOG ("\tThematic map: " << thematicMapFname << "\n");
									INDENT LOG ("\tConfidence map: " << confidenceMapFname << "\n");

									LOG_LEAVE (0);
}

/*************************************************************************************/
void UC_7_a_raisedVigilance () {
									LOG_ENTER (0, "\n");
									INDENT LOG ("************************************************************************************");  timeStamp(ts); LOG (ts << endl);
									INDENT LOG ("Use case 7.a - Model variations - raised vigilance:\n");
									INDENT LOG ("--------------------------------------------------\n");

	// Use a ViewSet object to automate the splitting and generation of orderings
	ViewSet *vs = new ViewSet ("bosViewSet", bosGt, 2, 5);
	View *v0 = vs->getView(0);
	View *v1 = vs->getView(1);

	//Create a couple more ViewSets to hold the training and testing views.
	ViewSet *vsTr = new ViewSet ("Training");
	vsTr->addView(v0);
	ViewSet *vsTe = new ViewSet ("Testing");
	vsTe->addView(v1);

	Classifier *cls = new Classifier (*(new string ("Artmap")), dimList, 5, *classNames);
	PropertyList *modelParams = cls->getModelParams();

									INDENT LOG ("Raising ARTMAP vigilance parameter to 0.9\n");
	modelParams->addParam("RhoBar", "0.9");
	cls->registerParams();

									INDENT LOG ("Training with half the data\n");
	cls->train(vsTr);
									INDENT LOG ("Training completed in " << cls->getTrainTime() << " seconds\n");
	cls->test(vsTe);
									INDENT LOG ("Testing completed in " << cls->getTestTime() << " seconds, " << cls->getPercentCorrect() << "% correct\n");
									INDENT LOG ("ARTMAP category nodes required: ");

	int num = cls->getNumClassifiers();

	for (int i = 0; i < num; i++) { LOG (cls->getClassifier(i)->getInt("f2Nodes") << " "); }
									LOG ("\n");

	ConfusionMatrix *cm = cls->getConfusionMatrix();
	cout << *(cm->charPlot());

	delete vs;
	delete v0;
	delete v1;
	delete cls;
	delete vsTr;
	delete vsTe;
	delete cm;

									LOG_LEAVE (0);
}

/*************************************************************************************/
void UC_7_b_differentInputFeatures () {
																	LOG_ENTER (0, "\n");
									INDENT LOG ("************************************************************************************");  timeStamp(ts); LOG (ts << endl);
									INDENT LOG ("Use case 7.b - Model variations - varying the input features:\n");
									INDENT LOG ("------------------------------------------------------------\n");

	// Use a ViewSet object to automate the splitting and generation of orderings
	ViewSet *vs = new ViewSet ("bosViewSet", bosGt, 2, 5);
	View *v0 = vs->getView(0);
	View *v1 = vs->getView(1);

	//Create a couple more ViewSets to hold the training and testing views.
	ViewSet *vsTr = new ViewSet ("Training");
	vsTr->addView(v0);
	ViewSet *vsTe = new ViewSet ("Testing");
	vsTe->addView(v1);

	Classifier *cls = new Classifier (*(new string ("Artmap")), dimList, 5, *classNames);

	DimList *dimList2 = genFeatureRange ("halfTheFeatures", "d", 1, numFeatures/2);

									INDENT LOG ("Setting classifier to only use the first half of the available features\n");
	cls->setDimList(dimList2);

									INDENT LOG ("Training with half the data\n");
	cls->train(vsTr);
									INDENT LOG ("Training completed in " << cls->getTrainTime() << " seconds\n");
	cls->test(vsTe);
									INDENT LOG ("Testing completed in " << cls->getTestTime() << " seconds, " << cls->getPercentCorrect() << "% correct\n");

	ConfusionMatrix *cm = cls->getConfusionMatrix();
	cout << *(cm->charPlot());

	delete dimList2;
	delete vs;
	delete v0;
	delete v1;
	delete cls;
	delete vsTr;
	delete vsTe;
	delete cm;

									LOG_LEAVE (0);
}

/*************************************************************************************/
void UC_8_a_iterateTrainTest () {
									LOG_ENTER (0, "\n");
									INDENT LOG ("************************************************************************************");  timeStamp(ts); LOG (ts << endl);
									INDENT LOG ("Use case 8.a - Iterated training/testing:\n");
									INDENT LOG ("----------------------------------------\n");

	// Use a ViewSet object to automate the splitting and generation of orderings
	ViewSet *vs = new ViewSet ("bosViewSet", bosGt, 2, 1);
	View *v0 = vs->getView(0);
	View *v1 = vs->getView(1);

	//Create a couple more ViewSets to hold the training and testing views.
	ViewSet *vsTr = new ViewSet ("Training");
	vsTr->addView(v0);

	ViewSet *vsTe = new ViewSet ("Testing");
	vsTe->addView(v1);

	Classifier *cls = new Classifier (*(new string ("Artmap")), dimList, 1, *classNames);

									INDENT LOG ("Iterating train/test with increasing numbers of input features:\n\n");
									INDENT LOG ("Input        Train     Test     Percent   # of \n");
									INDENT LOG ("Range        Time (s)  Time (s) Correct   Nodes\n");
									INDENT LOG ("-----        --------  -------- --------  -----\n");
	for (int numInputs = 10; numInputs <= numFeatures; numInputs += 2) {
		DimList *dims = genFeatureRange ("", "d", 1, numInputs);

		cls->setDimList (dims);
		cls->setHeartbeat(false);
		
		cls->train(vsTr);
		cls->test(vsTe);
	
									INDENT LOG ("[d1..d" << numInputs << "]    " 
			<< std::fixed << std::setw(4) << std::setprecision(2) << cls->getTrainTime () << "      "
			<< std::fixed << std::setw(4) << std::setprecision(2) << cls->getTestTime() << "    "
			<< std::fixed << std::setw(5) << std::setprecision(1) << cls->getPercentCorrect() << "%     " 
			<< std::fixed << std::setw(2) << cls->getClassifier(0)->getInt("f2Nodes") << "\n");

		delete dims;
	}

	delete vs;
	delete v0;
	delete v1;
	delete cls;
	delete vsTr;
	delete vsTe;

									LOG_LEAVE (0);
}

/*************************************************************************************/
void UC_8_b_iterateCvTrainTest () {
									LOG_ENTER (0, "\n");
									INDENT LOG ("************************************************************************************");  timeStamp(ts); LOG (ts << endl);
									INDENT LOG ("Use case 8.b - Iterated training/testing (five-fold cross-validation):\n");
									INDENT LOG ("---------------------------------------------------------------------\n");

	// Use a ViewSet object to automate the splitting and generation of orderings
	ViewSet *vs = new ViewSet ("bosViewSet", bosGt, 5, 1);

	Classifier *cls = new Classifier (*(new string ("Artmap")), dimList, 1, *classNames);

									INDENT LOG ("Iterating train/test with increasing numbers of input features: (cross-validated)\n\n");
									INDENT LOG ("Input        Train     Test     Percent   # of \n");
									INDENT LOG ("Range        Time (s)  Time (s) Correct   Nodes\n");
									INDENT LOG ("-----        --------  -------- --------  -----\n");
	for (int numInputs = 10; numInputs <= numFeatures; numInputs += 2) {
		DimList *dims = genFeatureRange ("", "d", 1, numInputs);

		cls->setDimList (dims);
		cls->setHeartbeat(false);
		
		cls->CvTrain(vs);
		cls->CvTest(vs);
	
									INDENT LOG ("[d1..d" << numInputs << "]    " 
			<< std::fixed << std::setw(4) << std::setprecision(2) << cls->getTrainTime () << "      "
			<< std::fixed << std::setw(4) << std::setprecision(2) << cls->getTestTime() << "    "
			<< std::fixed << std::setw(5) << std::setprecision(1) << cls->getPercentCorrect() << "%     " 
			<< std::fixed << std::setw(2) << cls->getClassifier(0)->getInt("f2Nodes") << "\n");

		delete dims;
	}

	delete vs;
	delete cls;

									LOG_LEAVE (0);
}

/*************************************************************************************/
void UC_9_genDecisionSurface () {
									LOG_ENTER (0, "\n");
									INDENT LOG ("************************************************************************************");  timeStamp(ts); LOG (ts << endl);
									INDENT LOG ("Use case 9 - Generate a decision surface:\n");
									INDENT LOG ("----------------------------------------\n");

	ViewSet *vs = new ViewSet ("bosViewSet", bosGt, 1, 1);

	Classifier *cls = new Classifier (*(new string ("Artmap")), dimList, 1, *classNames);
	DimList *dl = genFeatureRange ("", "d", 1, 2);
	cls->setDimList (dl);

	ofstream *ofsDecisionSurf = new ofstream (decisionSurfFname, ios::out | ios::binary);
	if (!(*ofsDecisionSurf)) MSG_EXCEPTION ("UC_9_genDecisionSurf() - Failed to open decision surface output file " << decisionSurfFname);

	cls->setPredPpmStream(*ofsDecisionSurf);
									INDENT LOG ("Training with only " << dl->getDimNames() << " as inputs, and generating a decision surface\n");
	cls->train(vs);

									INDENT LOG ("Done.\n\n");

	ofstream *ofsBostonLegend = new ofstream (bostonLegendFname, ios::out | ios::binary);
	if (!(*ofsBostonLegend)) MSG_EXCEPTION ("UC_9_genDecisionSurface() - Failed to open decision surface output file " << bostonLegendFname);

	Palette *palette = Palette::Instance();
	palette->writeLegend(*ofsBostonLegend, *classNames, true);

									INDENT LOG ("Finished output:\n");
									INDENT LOG ("\tDecision surface: " << decisionSurfFname << " (compare to plot in Use Case 1)\n");
									INDENT LOG ("\tClass legend:     " << bostonLegendFname << " (corresponds to class name order)\n");

									LOG_LEAVE (0);
}

/*************************************************************************************/
void UC_10_saveLoadClassifier () {
									LOG_ENTER (0, "\n");
									INDENT LOG ("************************************************************************************");  timeStamp(ts); LOG (ts << endl);
									INDENT LOG ("Use case 10 - Save/Load a classifier:\n");
									INDENT LOG ("------------------------------------\n");

									INDENT LOG ("Saving " << clsAllGt->getNumClassifiers() << " classifier model instances to files\n");
									INDENT LOG ("\tCategory nodes: ");
									for (int i = 0; i < clsAllGt->getNumClassifiers(); i++) {
										LOG (clsAllGt->getClassifier(i)->getInt("f2Nodes") << " ");
									}
									LOG ("\n");
	string fNameTemplate ("savedClassifier");
	clsAllGt->save(fNameTemplate);

									INDENT LOG ("Creating a new classifier and loading if with saved information\n");
	Classifier *newCls = new Classifier (string ("Artmap"), dimList, 5, *classNames);
	string specialRequest("");
	newCls->load(fNameTemplate, specialRequest);
									INDENT LOG ("New classifier model instances loaded from saved information\n");
									INDENT LOG ("\tCategory nodes: ");
									for (int i = 0; i < newCls->getNumClassifiers(); i++) {
										LOG (newCls->getClassifier(i)->getInt("f2Nodes") << " ");
									}
									LOG ("\n");

									LOG_LEAVE (0);
}

/*************************************************************************************/
void UC_11_fileOutputRequests () {
									LOG_ENTER (0, "\n");
									INDENT LOG ("************************************************************************************");  timeStamp(ts); LOG (ts << endl);
									INDENT LOG ("Use case 11 - Requests for outputs to files:\n");
									INDENT LOG ("-------------------------------------------\n");

									INDENT LOG ("Creating small subset of five pixels\n");
	View *v = new View ("smallSampleOfBostonData", bosImg);
	OrderedSubset *o = new OrderedSubset ("smallSample");
  o->add(100); o->add(200); o->add(300); o->add(400); o->add(500); 
	v->addOrdering (o);
	smallSubset = new ViewSet ("smallSubset");
	smallSubset->addView(v);

	ofstream *ofsOutputClass = new ofstream (outputClassFname, ios::out);
	if (!(*ofsOutputClass)) MSG_EXCEPTION ("UC_11_fileOutputRequests() - Failed to open class output file " << outputClassFname);
	ofstream *ofsOutputVector = new ofstream (outputVectorFname, ios::out);
	if (!(*ofsOutputVector)) MSG_EXCEPTION ("UC_11_fileOutputRequests() - Failed to open vector output file " << outputVectorFname);

	clsAllGt->setOutputClassStream(*ofsOutputClass);
	clsAllGt->setOutputVectorStream(*ofsOutputVector);

									INDENT LOG ("Testing small subset, sending outputs to file\n");
	clsAllGt->test(smallSubset);

									INDENT LOG ("Finished output:\n");
									INDENT LOG ("\tClass outputs:  " << outputClassFname << "\n");
									INDENT LOG ("\tVector outputs: " << outputVectorFname << "\n");

									LOG_LEAVE (0);
}

/*************************************************************************************/
void UC_12_cappingClasses () {
									LOG_ENTER (0, "\n");
									INDENT LOG ("************************************************************************************");  timeStamp(ts); LOG (ts << endl);
									INDENT LOG ("Use case 12 - Capping classes during training:\n");
									INDENT LOG ("---------------------------------------------\n");

	ViewSet *vs = new ViewSet ("bosViewSet", bosGt, 5, 5);

	Classifier *cls = new Classifier (*(new string ("Artmap")), dimList, 5, *classNames);

	cls->setHeartbeat(false);
									INDENT LOG ("Training (cross-validated)\n");
	cls->CvTrain(vs);

	LOG ("\nTraining samples, by class:\n");
	cout << *(cls->getClassStats());
									INDENT LOG ("Training completed in " << cls->getTrainTime() << " seconds\n");
	cls->CvTest(vs);
									INDENT LOG ("Testing completed in " << cls->getTestTime() << " seconds, " << cls->getPercentCorrect() << "% correct\n");

	ConfusionMatrix *cm = cls->getConfusionMatrix();
	cout << *(cm->charPlot());

	delete cls;

	cls = new Classifier (*(new string ("Artmap")), dimList, 5, *classNames);

	cls->setHeartbeat(false);
									INDENT LOG ("Training (cross-validated, with class cap = 250)\n");
	cls->setClassCap(150);
	cls->CvTrain(vs);

	LOG ("\nTraining samples, by class:\n");
	cout << *(cls->getClassStats());
									INDENT LOG ("Training completed in " << cls->getTrainTime() << " seconds\n");
	cls->setClassCap(0);
	cls->CvTest(vs);
									INDENT LOG ("Testing completed in " << cls->getTestTime() << " seconds, " << cls->getPercentCorrect() << "% correct\n");

	cm = cls->getConfusionMatrix();
	cout << *(cm->charPlot());

	delete vs;
	delete cls;
	delete cm;

									LOG_LEAVE (0);
}

/*************************************************************************************/
void UC_13_classerLogging () {
									LOG_ENTER (0, "\n");
									INDENT LOG ("************************************************************************************");  timeStamp(ts); LOG (ts << endl);
									INDENT LOG ("Use case 13 - Turning on Classer logging:\n");
									INDENT LOG ("----------------------------------------\n");


	ViewSet *vs = new ViewSet ("bosViewSet", bosGt, 1, 1);

	Classifier *cls = new Classifier (*(new string ("Artmap")), dimList, 1, *classNames);
	cls->train(vs);

									INDENT LOG ("===================================================================================================\n");
									INDENT LOG ("Testing small data subset with log level 3\n");
	Logger::setLogLevel(3);
	cls->test(smallSubset);

									INDENT LOG ("===================================================================================================\n");
									INDENT LOG ("Testing small data subset with log level 4\n");
	Logger::setLogLevel(4);
	cls->test(smallSubset);

									INDENT LOG ("===================================================================================================\n");
									INDENT LOG ("Testing small data subset with log level 5\n");
	Logger::setLogLevel(5);
	cls->test(smallSubset);

	delete cls;
									INDENT LOG ("===================================================================================================\n");
	Logger::setLogLevel(0);
									LOG_LEAVE (0);
}

/*************************************************************************************/
void UC_14_confusionMatrices () {
									LOG_ENTER (0, "\n");
									INDENT LOG ("************************************************************************************");  timeStamp(ts); LOG (ts << endl);
									INDENT LOG ("Use case 14 - Generating confusion matrices:\n");
									INDENT LOG ("-------------------------------------------\n");

	ViewSet *vs = new ViewSet ("bosViewSet", bosGt, 5, 5);

	Classifier *cls = new Classifier (*(new string ("Artmap")), dimList, 5, *classNames);
	cls->setHeartbeat(false);

									INDENT LOG ("Training (cross-validated)\n");
	cls->CvTrain(vs);
									INDENT LOG ("Training completed in " << cls->getTrainTime() << " seconds\n");

	ofstream *ofsConfusionMatrix = new ofstream (confusionMatrixFname, ios::out | ios::binary);
	if (!(*ofsConfusionMatrix)) MSG_EXCEPTION ("UC_15_confusionMatrices() - Failed to open confusion matrix output file " << confusionMatrixFname);
	cls->setCmatPgmStream(*ofsConfusionMatrix);
	cls->CvTest(vs);
									INDENT LOG ("Testing completed in " << cls->getTestTime() << " seconds, " << cls->getPercentCorrect() << "% correct\n");

									INDENT LOG ("Generating text-based confusion matrix:\n");
	ConfusionMatrix *cm = cls->getConfusionMatrix();
	cout << *(cm->charPlot());
	
									LOG ("\n");
									INDENT LOG ("Compare to graphical confusion matrix, directed to -> " << confusionMatrixFname << "\n");

	delete vs;
	delete cls;
	delete cm;

									LOG_LEAVE (0);
}

/*************************************************************************************/
void UC_15_randomSeed () {
									LOG_ENTER (0, "\n");
									INDENT LOG ("************************************************************************************");  timeStamp(ts); LOG (ts << endl);
									INDENT LOG ("Use case 15 - Setting the random seed:\n");
									INDENT LOG ("-------------------------------------\n");

									INDENT LOG ("Running with a random seed of 1\n");
	srand (1);
		// Use a ViewSet object to automate the splitting and generation of orderings
	ViewSet *vs = new ViewSet ("bosViewSet", bosGt, 2, 1);
	View *v0 = vs->getView(0);
	View *v1 = vs->getView(1);
	// Randomize the first ordering - by default, it's in order
	v0->getOrdering(0)->permuteEntries();
	v1->getOrdering(0)->permuteEntries();

	//Create a couple more ViewSets to hold the training and testing views.
	ViewSet *vsTr = new ViewSet ("Training");
	vsTr->addView(v0);
	ViewSet *vsTe = new ViewSet ("Testing");
	vsTe->addView(v1);

	Classifier *cls = new Classifier (*(new string ("Artmap")), dimList, 1, *classNames);

									INDENT LOG ("Training with half the data\n");
	cls->train(vsTr);
									INDENT LOG ("Training completed in " << cls->getTrainTime() << " seconds\n");
	cls->test(vsTe);
									INDENT LOG ("Testing completed in " << cls->getTestTime() << " seconds, " << cls->getPercentCorrect() << "% correct\n");
									INDENT LOG ("ARTMAP category nodes required: ");

	ConfusionMatrix *cm = cls->getConfusionMatrix();
	cout << *(cm->charPlot());

	delete vs;
	delete v0;
	delete v1;
	delete cls;
	delete vsTr;
	delete vsTe;
	delete cm;

									INDENT LOG ("Running with a random seed of 4\n");
	srand (4);
	// Use a ViewSet object to automate the splitting and generation of orderings
	vs = new ViewSet ("bosViewSet", bosGt, 2, 1);
	v0 = vs->getView(0);
	v1 = vs->getView(1);
	// Randomize the first ordering - by default, it's in order
	v0->getOrdering(0)->permuteEntries();
	v1->getOrdering(0)->permuteEntries();

	//Create a couple more ViewSets to hold the training and testing views.
	vsTr = new ViewSet ("Training");
	vsTr->addView(v0);
	vsTe = new ViewSet ("Testing");
	vsTe->addView(v1);

	cls = new Classifier (*(new string ("Artmap")), dimList, 1, *classNames);

									INDENT LOG ("Training with half the data\n");
	cls->train(vsTr);
									INDENT LOG ("Training completed in " << cls->getTrainTime() << " seconds\n");
	cls->test(vsTe);
									INDENT LOG ("Testing completed in " << cls->getTestTime() << " seconds, " << cls->getPercentCorrect() << "% correct\n");

	cm = cls->getConfusionMatrix();
	cout << *(cm->charPlot());

	delete vs;
	delete v0;
	delete v1;
	delete cls;
	delete vsTr;
	delete vsTe;
	delete cm;



									LOG_LEAVE (0);
}

/*************************************************************************************/
int main(int argc, char *argv[]) {
	try {
									INDENT LOG ("ClasserTester - A driver for Classer Use Cases\n");
									INDENT LOG ("----------------------------------------------\n");
									INDENT LOG ("  See 'Classer API Use Cases' document for details\n");

									LOG_ENTER (0, "");
	
		UC_0_defineClassesAndFeatures();

		// First see if the binary fine already exists
		// If so, use it directly
		ifstream *imgBinIfs = new ifstream (bostonImgFname_bin, ios::in | ios::binary);
		if (!(*imgBinIfs)) { // If not, load raw data and preprocess
		  UC_0_preprocess();
		} else {
									INDENT LOG ("Preprocessed data exists, using it directly\n");	
		}
	
		UC_1_loadImage();
		UC_2_defineGroundTruth();
		UC_3_trainTest();
		UC_4_CVtrainTest();
		UC_5_trainAllGt();
		UC_6_genThematicMap();
		UC_7_a_raisedVigilance();
		UC_7_b_differentInputFeatures();
		UC_8_a_iterateTrainTest();
		UC_8_b_iterateCvTrainTest();
		UC_9_genDecisionSurface();
		UC_10_saveLoadClassifier();
		UC_11_fileOutputRequests();
		UC_12_cappingClasses();
		UC_13_classerLogging();
		UC_14_confusionMatrices();
		UC_15_randomSeed();

									timeStamp(ts);
									LOG ("\n"); 
									INDENT LOG ("Time: " << ts << endl);

									LOG_LEAVE (0);
	} catch (MsgException &e) {
		cout << "Error: " << e.getMsg() << endl; 
	}
	return 0;
}